@@ -23,49 +23,14 @@
 #include <string.h>
 #include <stdio.h>
 
-#include "dconf-rebuilder.h"
-
-static const GDBusArgInfo name_arg = { -1, "name", "s" };
-static const GDBusArgInfo names_arg = { -1, "names", "as" };
-static const GDBusArgInfo serial_arg = { -1, "serial", "t" };
-static const GDBusArgInfo locked_arg = { -1, "locked", "b" };
-static const GDBusArgInfo value_arg = { -1, "value", "av" };
-static const GDBusArgInfo values_arg = { -1, "values", "a(sav)" };
-
-static const GDBusArgInfo *write_inargs[] = { &name_arg, &value_arg, NULL };
-static const GDBusArgInfo *write_outargs[] = { &serial_arg, NULL };
-
-static const GDBusArgInfo *merge_inargs[] = { &name_arg, &values_arg, NULL };
-static const GDBusArgInfo *merge_outargs[] = { &serial_arg, NULL };
-
-static const GDBusArgInfo *gsd_inargs[] = { NULL };
-static const GDBusArgInfo *gsd_outargs[] = { &name_arg, NULL };
-
-static const GDBusMethodInfo write_method = {
-  -1, "Write",         (gpointer) write_inargs, (gpointer) write_outargs };
-static const GDBusMethodInfo merge_method = {
-  -1, "Merge",         (gpointer) merge_inargs, (gpointer) merge_outargs };
-static const GDBusMethodInfo gsd_method = {
-  -1, "GetSessionDir", (gpointer) gsd_inargs,   (gpointer) gsd_outargs };
-
-static const GDBusMethodInfo *writer_methods[] = {
-  &write_method, &merge_method, &gsd_method, NULL
-};
-
-static const GDBusInterfaceInfo writer_interface = {
-  -1, "ca.desrt.dconf.Writer", (gpointer) writer_methods
-};
-
-typedef struct
-{
-  GMainLoop *loop;
-  guint64 serial;
-  gchar *path;
-} DConfWriter;
+#include "dconf-interfaces.h"
+#include "dconf-writer.h"
+#include "dconf-state.h"
 
 static void
 emit_notify_signal (GDBusConnection  *connection,
-                    guint64           serial,
+                    DConfWriter      *writer,
+                    const gchar      *tag,
                     const gchar      *prefix,
                     const gchar     **keys,
                     guint             n_keys)
@@ -73,6 +38,7 @@ emit_notify_signal (GDBusConnection  *connection,
   GVariantBuilder builder;
   GVariant *items;
   gchar *path;
+  gchar *obj;
 
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("as"));
 
@@ -115,15 +81,18 @@ emit_notify_signal (GDBusConnection  *connection,
   else
     path = g_strdup (prefix);
 
+  obj = g_strjoin (NULL, "/ca/desrt/dconf/Writer/",
+                   dconf_writer_get_name (writer), NULL);
   g_dbus_connection_emit_signal (connection, NULL, "/",
                                  "ca.desrt.dconf.Writer", "Notify",
-                                 g_variant_new ("(ts@as)",
-                                                serial, path, items),
+                                 g_variant_new ("(s@ass)",
+                                                path, items, tag),
                                  NULL);
   g_free (path);
+  g_free (obj);
 }
 
-static GVariant *
+static void
 unwrap_maybe (GVariant **ptr)
 {
   GVariant *array, *child;
@@ -150,6 +119,9 @@ method_call (GDBusConnection       *connection,
              gpointer               user_data)
 {
   DConfWriter *writer = user_data;
+  DConfState *state;
+
+  state = dconf_writer_get_state (writer);
 
   if (strcmp (method_name, "Write") == 0)
     {
@@ -158,8 +130,9 @@ method_call (GDBusConnection       *connection,
       const gchar *key;
       gsize key_length;
       GVariant *value;
-      guint64 serial;
       GVariant *none;
+      gchar *path;
+      gchar *tag;
 
       g_variant_get (parameters, "(@s@av)", &keyvalue, &value);
       key = g_variant_get_string (keyvalue, &key_length);
@@ -186,40 +159,44 @@ method_call (GDBusConnection       *connection,
           return;
         }
 
-      if (!dconf_rebuilder_rebuild (writer->path, "", &key,
-                                    &value, 1, &error))
+      if (!dconf_writer_write (writer, key, value, &error))
         {
           g_dbus_method_invocation_return_gerror (invocation, error);
           g_error_free (error);
           return;
         }
 
-      serial = writer->serial++;
+      tag = dconf_state_get_tag (state);
       g_dbus_method_invocation_return_value (invocation,
-                                             g_variant_new ("(t)", serial));
+                                             g_variant_new ("(s)", tag));
       none = g_variant_new_array (G_VARIANT_TYPE_STRING, NULL, 0);
-      g_dbus_connection_emit_signal (connection, NULL, "/",
+      path = g_strjoin (NULL, "/ca/desrt/dconf/Writer/",
+                        dconf_writer_get_name (writer), NULL);
+      g_dbus_connection_emit_signal (connection, NULL, path,
                                      "ca.desrt.dconf.Writer", "Notify",
-                                     g_variant_new ("(ts@as)",
-                                                    serial, key, none),
+                                     g_variant_new ("(s@ass)",
+                                                    key, none, tag),
                                      NULL);
+      g_free (path);
+      g_free (tag);
     }
-  else if (strcmp (method_name, "Merge") == 0)
+
+  else if (strcmp (method_name, "WriteMany") == 0)
     {
       GError *error = NULL;
       const gchar *prefix;
       GVariantIter *iter;
       const gchar **keys;
       GVariant **values;
-      guint64 serial;
       gsize length;
+      gchar *tag;
       gint i = 0;
       gint j;
 
       g_variant_get (parameters, "(&sa(sav))", &prefix, &iter);
       length = g_variant_iter_n_children (iter);
 
-      keys = g_new (const gchar *, length);
+      keys = g_new (const gchar *, length + 1);
       values = g_new (GVariant *, length);
       while (g_variant_iter_next (iter, "(&s@av)", &keys[i], &values[i]))
         {
@@ -247,19 +224,17 @@ method_call (GDBusConnection       *connection,
       g_variant_iter_free (iter);
       keys[i] = NULL;
 
-      if (!dconf_rebuilder_rebuild (writer->path, prefix, keys,
-                                    values, i, &error))
+      if (!dconf_writer_write_many (writer, prefix, keys, values, i, &error))
         {
           g_dbus_method_invocation_return_gerror (invocation, error);
           g_error_free (error);
           return;
         }
 
-      serial = writer->serial++;
-
+      tag = dconf_state_get_tag (state);
       g_dbus_method_invocation_return_value (invocation,
-                                             g_variant_new ("(t)", serial));
-      emit_notify_signal (connection, serial, prefix, keys, i);
+                                             g_variant_new ("(s)", tag));
+      emit_notify_signal (connection, writer, tag, prefix, keys, i);
 
       for (j = 0; j < i; j++)
         if (values[j] != NULL)
@@ -267,22 +242,148 @@ method_call (GDBusConnection       *connection,
 
       g_free (values);
       g_free (keys);
+      g_free (tag);
     }
+
+  else if (strcmp (method_name, "SetLocked") == 0)
+    {
+      GError *error = NULL;
+      const gchar *name;
+      gboolean locked;
+
+      g_variant_get (parameters, "(&sb)", &name, &locked);
+
+      if (!dconf_writer_set_lock (writer, name, locked, &error))
+        {
+          g_dbus_method_invocation_return_gerror (invocation, error);
+          g_error_free (error);
+          return;
+        }
+
+      g_dbus_method_invocation_return_value (invocation, NULL);
+    }
+
   else
     g_assert_not_reached ();
 }
 
+static GVariant *
+writer_info_get_property (GDBusConnection  *connection,
+                          const gchar      *sender,
+                          const gchar      *object_path,
+                          const gchar      *interface_name,
+                          const gchar      *property_name,
+                          GError          **error,
+                          gpointer          user_data)
+{
+  DConfState *state = user_data;
+
+  return g_variant_new_string (state->shm_dir);
+}
+
+static const GDBusInterfaceVTable *
+subtree_dispatch (GDBusConnection *connection,
+                  const gchar     *sender,
+                  const gchar     *object_path,
+                  const gchar     *interface_name,
+                  const gchar     *node,
+                  gpointer        *out_user_data,
+                  gpointer         user_data)
+{
+  DConfState *state = user_data;
+
+  if (strcmp (interface_name, "ca.desrt.dconf.Writer") == 0)
+    {
+      static const GDBusInterfaceVTable vtable = {
+        method_call, NULL, NULL
+      };
+      static GHashTable *writer_table;
+      DConfWriter *writer;
+
+      if (node == NULL)
+        return NULL;
+
+      if G_UNLIKELY (writer_table == NULL)
+        writer_table = g_hash_table_new (g_str_hash, g_str_equal);
+
+      writer = g_hash_table_lookup (writer_table, node);
+
+      if G_UNLIKELY (writer == NULL)
+        {
+          writer = dconf_writer_new (state, node);
+          g_hash_table_insert (writer_table, g_strdup (node), writer);
+        }
+
+      *out_user_data = writer;
+
+      return &vtable;
+    }
+
+  else if (strcmp (interface_name, "ca.desrt.dconf.WriterInfo") == 0)
+    {
+      static const GDBusInterfaceVTable vtable = {
+        NULL, writer_info_get_property, NULL
+      };
+
+      *out_user_data = state;
+      return &vtable;
+    }
+
+  else
+    return NULL;
+}
+
+static gchar **
+subtree_enumerate (GDBusConnection *connection,
+                   const gchar     *sender,
+                   const gchar     *object_path,
+                   gpointer         user_data)
+{
+  return dconf_writer_list_existing ();
+}
+
+static GDBusInterfaceInfo **
+subtree_introspect (GDBusConnection *connection,
+                    const gchar     *sender,
+                    const gchar     *object_path,
+                    const gchar     *node,
+                    gpointer         user_data)
+{
+  /* The root node supports only the info iface */
+  if (node == NULL)
+    {
+      const GDBusInterfaceInfo *interfaces[] = {
+        &ca_desrt_dconf_WriterInfo, NULL
+      };
+
+      return g_memdup (interfaces, sizeof interfaces);
+    }
+  else
+    {
+      const GDBusInterfaceInfo *interfaces[] = {
+        &ca_desrt_dconf_WriterInfo, &ca_desrt_dconf_Writer, NULL
+      };
+
+      return g_memdup (interfaces, sizeof interfaces);
+    }
+}
+
 static void
 bus_acquired (GDBusConnection *connection,
               const gchar     *name,
               gpointer         user_data)
 {
-  static const GDBusInterfaceVTable interface_vtable = { method_call };
-  DConfWriter *writer = user_data;
+  static GDBusSubtreeVTable vtable = {
+    subtree_enumerate, subtree_introspect, subtree_dispatch
+  };
+  DConfState *state = user_data;
+  GDBusSubtreeFlags flags;
 
-  g_dbus_connection_register_object (connection, "/",
-                                     &writer_interface, &interface_vtable,
-                                     writer, NULL, NULL);
+  dconf_state_set_id (state, g_dbus_connection_get_unique_name (connection));
+
+  flags = G_DBUS_SUBTREE_FLAGS_DISPATCH_TO_UNENUMERATED_NODES;
+  g_dbus_connection_register_subtree (connection, "/ca/desrt/dconf/Writer",
+                                      &vtable, flags, state, NULL, NULL);
 }
 
 static void
@@ -290,10 +391,6 @@ name_acquired (GDBusConnection *connection,
                const gchar     *name,
                gpointer         user_data)
 {
-  DConfWriter *writer = user_data;
-
-  writer->serial = time (NULL);
-  writer->serial <<= 32;
 }
 
 static void
@@ -301,31 +398,29 @@ name_lost (GDBusConnection *connection,
            const gchar     *name,
            gpointer         user_data)
 {
-  DConfWriter *writer = user_data;
-
-  if (writer->serial != 0)
-    g_critical ("dbus signaled that we lost our "
-                "name but it wrong wrong to do so");
-  else
-    g_critical ("unable to acquire name: '%s'", name);
-
-  g_main_loop_quit (writer->loop);
+  g_error ("unable to acquire name: '%s'", name);
 }
 
 int
 main (void)
 {
-  DConfWriter writer = {  };
+  DConfState state;
+  GBusType type;
 
   g_type_init ();
+  dconf_state_init (&state);
 
-  writer.loop = g_main_loop_new (NULL, FALSE);
-  writer.path = g_build_filename (g_get_user_config_dir (), "dconf", NULL);
+  if (state.is_session)
+    type = G_BUS_TYPE_SESSION;
+  else
+    type = G_BUS_TYPE_SYSTEM;
 
-  g_bus_own_name (G_BUS_TYPE_SESSION, "ca.desrt.dconf", 0,
-                  bus_acquired, name_acquired, name_lost, &writer, NULL);
+  g_bus_own_name (type, "ca.desrt.dconf", G_BUS_NAME_OWNER_FLAGS_NONE,
+                  bus_acquired, name_acquired, name_lost, &state, NULL);
 
-  g_main_loop_run (writer.loop);
+  g_main_loop_run (state.main_loop);
+
+  dconf_state_destroy (&state);
 
   return 0;
 }
